package provide DS9 1.0

# Open

proc OpenFile {} {
    global ds9
    LoadFits [OpenFileDialog $ds9(top) ds9(lastFitsName)] ""
}

proc OpenArray {} {
    global ds9
    LoadArray [OpenFileDialog $ds9(top) ds9(lastArrayName)]
}

proc OpenMosaicFile {} {
    global ds9
    LoadFits [OpenFileDialog $ds9(top) ds9(lastMosaicName)] mosaic
}

proc OpenURL {} {
    global ds9
    LoadFits [URLDialog ds9(lastURL)] ""
}

proc OpenMosaicURL {} {
    global ds9
    LoadFits [URLDialog ds9(lastURL)] mosaic
}

proc OpenMosaicExtFile {} {
    global ds9
    LoadMosaicExtFile [OpenFileDialog $ds9(top) ds9(lastMosaicName)]
}

proc OpenMosaicFiles {} {
    global ds9
    LoadMosaicFiles [OpenFileDialog $ds9(top) ds9(lastMosaicName)]
}

# Load

proc LoadFits {url mode} {
    if {$url != ""} {
	if {! [regexp -nocase {^([A-Za-z]+)?(:)?(/)*([^/:]+)(:([0-9]+))?(/.*)?$} \
		   $url x proto ask slash host y port srvurl]} {
	    Error "Unable to parse URL $url"
	    return
	}

	switch $proto {
	    http {LoadHTTP $url "/tmp/[file tail $url]" 0 $mode {} {}}
	    ftp {LoadFTP $url $mode}
	    file {LoadFile "$slash$host$y$port$srvurl" $mode}
	    default {LoadFile $url $mode}
	}
    }
}

proc LoadFile {fileName mode} {
    global current

    StartLoad

    if {$fileName != ""} {
	global loadParam

	set loadParam(type) mmap
	set loadParam(mode) $mode
	set loadParam(filetype) fits
	ConvertFile $fileName
	ProcessLoad

	ResetFrameColorbar
    }

    FinishLoad
}

proc LoadArray {fileName} {
    global current

    StartLoad

    if {$fileName != ""} {
	global loadParam

	set loadParam(type) mmap
	set loadParam(mode) ""
	set loadParam(filetype) array
	ConvertFile $fileName
	ProcessLoad

	ResetFrameColorbar
    }

    Finishload
}

proc LoadFTP {url mode} {
    global current

    StartLoad

    regexp -nocase {^([A-Za-z]+)?(://)?([^/:]+)(:([0-9]+))?(/.*)?$} $url \
	x proto dummy host y port srvurl

    if [FTP::Open $host "ftp" "ds9@"] {
	set tmp "/tmp/[file tail $url]"
	FTP::Type binary
	if [FTP::Get $srvurl $tmp] {
	    global loadParam

	    # alloc it because we are going to delete it after load
	    set loadParam(type) alloc
	    set loadParam(mode) $mode
	    set loadParam(filetype) fits
	    ConvertFile $tmp
	    ProcessLoad

	    ResetFrameColorbar
	}
	if [file exists $tmp] {
	    catch {file delete -force $tmp}
	}
	FTP::Close
    }

    FinishLoad
}

proc LoadHTTP {url fn savefn mode progress query} {
    global loadParam
    global http

    StartLoad

    set ch [open $fn w]

    set cmd "http::geturl $url -channel $ch"
    if {$progress != ""} {
	set cmd "$cmd -progress $progress"
    }
    if {$query != ""} {
	set cmd "$cmd -query $query"
    }
    set token [eval $cmd]

    if {$http(log)} {
	upvar 0 $token t
	DisplayLog "$t(http)\n"
	DisplayLog "$t(type)\n"
	DisplayLog "$t(totalsize)\n"
	DisplayLog [BreakUp $t(meta)]
	DisplayLog "$t(status)\n"
    }

    # alloc it because we are going to delete it after load
    set loadParam(type) alloc
    set loadParam(mode) $mode
    set loadParam(filetype) fits
    ConvertFile $fn
    ProcessLoad

    catch {close $ch}

    ResetFrameColorbar

    if {!$savefn} {
	if {[file exists $fn]} {
	    catch {file delete -force $fn}
	}
    }

    FinishLoad
}

proc DisplayLog {item} {
    global ftpDialog
    global current

    set ftpDialog(top) .ftp
    set ftpDialog(mb) .ftpMB
    set ftpDialog(txt) "$item"

    SimpleTextDialog ftpDialog 80 40 "Mesage Log" append bottom
}

proc LoadShared {id size mode filter} {
    global loadParam
    global current

    StartLoad

    set loadParam(mode) $mode
    set loadParam(id) $id
    set loadParam(size) $size
    set loadParam(name) $filter
    set loadParam(filetype) fits
    ProcessLoad

    ResetFrameColorbar

    FinishLoad
}

proc LoadMosaicExtFile {fileName} {
    global current
    global loadParam

    StartLoad

    if {$fileName != ""} {
	set loadParam(type) mmap
	set loadParam(mode) "mosaic ext"
	set loadParam(filetype) fits
	ConvertFile $fileName
	ProcessLoad

	ResetFrameColorbar
    }

    FinishLoad
}

proc LoadMosaicFiles {fileName} {
    global current
    global loadParam
    global mosaicTypes

    StartLoad

    if {$fileName != ""} {
	$current(frame) unload

        foreach t $mosaicTypes {
            set fileNames [$t $fileName]
            if {$fileNames != ""} {
		break
	    }
        }
        if {$fileNames == ""} { 
	    set fileNames $fileName 
	}

        foreach fileName $fileNames {
	    set loadParam(type) mmap
	    set loadParam(mode) mosaic
	    set loadParam(filetype) fits
	    ConvertFile $fileName
	    ProcessLoad
        }

	ResetFrameColorbar
    }

    FinishLoad
}

# Support

proc BreakUp {str} {
    set r ""
    set l [string length $str]
    for {set i 0} {$i < $l} {incr i} {
	set c [string index $str $i]
	append r $c
	if {$c=="\}"} {
	    append r "\n"
	}
    }
    return $r
}

proc URLDialog {prevURL} {
    upvar $prevURL result

    if {[EntryDialog "URL" "Enter World Wide Web Location (URL):" 80 result]} {
	return $result
    } else {
	return ""
    }
}

# Header

proc DisplayHeaderMenu {} {
    global current

    set count [$current(frame) get fits count]
    switch $count {
	0 {return}
	1 {set which 1}
	default {
	    set which [SimpleListBox ".headerSelect" "Select Header" \
			   [$current(frame) get fits file names]]
	    if {$which==0} {
		return
	    }
	}
    }

    DisplayHeader $which
}

proc DisplayHeader {which} {
    global current
    set var "hd[string range $current(frame) end end]"

    eval "global $var"
    eval "set $var\(top) .$var"
    eval "set $var\(mb) .${var}MB"
    eval "set $var\(txt) \[\"$current(frame)\" get fits header \"$which\"\]"
    eval "SimpleTextDialog $var 80 40 \"$current(frame)\" insert top"

    # create a special text tag for keywords

    eval "\$$var\(text) tag configure keyword -font {courier} -foreground blue"
    
    # color tag keywords

    eval "set stop \[\$$var\(text) index end\]"
    for {set i 1.0} {$i<$stop} {set i [expr $i+1]} {
	eval "\$$var\(text) tag add keyword $i {$i +8 chars}"
    }
}

proc UpdateDisplayHeader {which} {
    set var "hd[string range $which end end]"

    eval "global $var"
    if {[eval "info exists $var\(top)"] && [eval "winfo exist \$$var\(top)"]} {
	DisplayHeaderMenu
    }
}

proc DestroyHeader {which} {
    set var "hd[string range $which end end]"

    eval "global $var"
    if {[eval "info exists $var\(top)"] && [eval "winfo exist \$$var\(top)"]} {
	eval "DestroySimpleTextDialog $var"
    }
}

proc UpdateFitsMenu {} {
    global current
    global buttons

    UpdateCoordMenu2 .menuBar.frame.match 0 0
    UpdateCoordMenu2 .menuBar.frame.crosshair 0 0

    if {$current(frame) != ""} {
	.menuBar.file entryconfig "Open..." -state normal
	.menuBar.file entryconfig "Open Array..." -state normal
	.menuBar.file entryconfig "Open URL..." -state normal
	.menuBar.file entryconfig "Open Mosaic Image File..." -state normal
	.menuBar.file entryconfig "Open Mosaic Image Multiple Files..." \
	    -state normal
	.menuBar.file entryconfig "Open Mosaic Segment..." -state normal
	.menuBar.file entryconfig "Open Mosaic Segment URL..." -state normal

	$buttons(name).file.open configure -state normal

	set count [$current(frame) get fits count]
	if {$count == 0} {
	    .menuBar.file entryconfig "Display Fits Header" -state disabled
	    .menuBar.frame entryconfig "Clear Frame" -state disabled

	    $buttons(name).file.clear configure -state disabled
	    $buttons(name).file.header configure -state disabled
	} else {
	    .menuBar.file entryconfig "Display Fits Header" -state normal
	    .menuBar.frame entryconfig "Clear Frame" -state normal

	    $buttons(name).file.clear configure -state normal
	    $buttons(name).file.header configure -state normal
	}

	.menuBar.frame entryconfig "Match Frames" -state normal
	.menuBar.frame entryconfig "Lock Crosshairs" -state normal

    } else {
	.menuBar.file entryconfig "Open..." -state disabled
	.menuBar.file entryconfig "Open Array..." -state disabled
	.menuBar.file entryconfig "Open URL..." -state disabled
	.menuBar.file entryconfig "Open Mosaic Image File..." -state disabled
	.menuBar.file entryconfig "Open Mosaic Image Multiple Files..." \
	    -state disabled
	.menuBar.file entryconfig "Open Mosaic Segment..." -state disabled
	.menuBar.file entryconfig "Open Mosaic Segment URL..." -state disabled
	$buttons(name).file.open configure -state normal

	.menuBar.file entryconfig "Display Fits Header" -state disabled
	$buttons(name).file.clear configure -state disabled
	$buttons(name).file.header configure -state disabled

	.menuBar.frame entryconfig "Clear Frame" -state disabled
	.menuBar.frame entryconfig "Match Frames" -state disabled
	.menuBar.frame entryconfig "Lock Crosshairs" -state disabled
    }
}

proc ProcessLoad {} {
    global current
    global loadParam

    switch $loadParam(type) {
	mmap {
	    if [catch {$current(frame) load $loadParam(filetype) \
			   $loadParam(mode) \
			   mmap $loadParam(name)}] {
		Error "Unable to load $loadParam(filetype) $loadParam(mode) Image $loadParam(name)"
	    }
	}
	alloc {
	    if [catch {$current(frame) load $loadParam(filetype) \
			   $loadParam(mode) \
			   alloc $loadParam(name)}] {
		Error "Unable to load $loadParam(filetype) $loadParam(mode) Image $loadParam(name)" 
	    }
	}
	shared {
	    if [catch {$current(frame) load $loadParam(filetype) \
			   $loadParam(mode) \
			   shared $loadParam(id) $loadParam(size) \
			   $loadParam(name)}] {
		Error "Unable to load shared $loadParam(filetype) $loadParam(mode) Image $loadParam(id)"
	    }
	}
	channel {
	    fconfigure $loadParam(channel) -translation binary
	    if [catch {$current(frame) load $loadParam(filetype) \
			   $loadParam(mode) \
			   channel $loadParam(channel) $loadParam(name)}] {
		Error "Unable to read $loadParam(filetype) $loadParam(mode) Image from channel $loadParam(channel)"
	    }
	    catch {close $loadParam(channel)}
	}
    }
}

proc StartLoad {} {
    SetWatchCursor
}

proc FinishLoad {} {
    global current

    GridUpdate
    ContourUpdate
    # do after updategrid and updatecontour
    UpdateDS9
    UnsetWatchCursor

    UpdateDisplayHeader $current(frame)
}

proc ConvertFile fileName {
    global ds9
    global loadParam

    # default

    set loadParam(name) $fileName

    foreach t $ds9(fileProcs) {
	if [$t $fileName] {
	    return
	}
    }
}

# File Types

proc StdinFits {fileName} {
    global loadParam

    if {$fileName == "-" || $fileName == "stdin" || $fileName == "STDIN"} {
	set loadParam(type) alloc
	set loadParam(name) "stdin"
	return 1
    }

    return 0
}

proc CompressFits {fileName} {
    global loadParam

    if { [regexp {(.*)\.Z($|\[)} $fileName matched root] } {
	if [catch {set ch [open "| uncompress < $root.Z " r]}] {
	    return 0
	}
	set loadParam(type) channel
	set loadParam(channel) $ch
	set loadParam(name) "$fileName"
	return 1
    }

    return 0
}

proc PackFits {fileName} {
    global loadParam

    if { [regexp {(.*)\.z($|\[)} $fileName matched root] } {
	if [catch {set ch [open "| pcat $root.z " r]}] {
	    return 0
	}
	set loadParam(type) channel
	set loadParam(channel) $ch
	set loadParam(name) "$fileName"
	return 1
    }

    return 0
}

proc GzipFits {fileName} {
    global loadParam

    if { [regexp {(.*)\.gz($|\[)} $fileName matched root] } {
	if [catch {set ch [open "| gunzip < $root.gz " r]}] {
	    return 0
	}
	set loadParam(type) channel
	set loadParam(channel) $ch
	set loadParam(name) "$fileName"
	return 1
    }

    return 0
}

proc ExternalFits {fileName} {
    global externalType
    global externalCmd
    global loadParam

    foreach a [array names externalType] {
	if { [regexp "(.*)\.$externalType($a)(.*)" $fileName matched] } {
	    regsub -all {\$filename} "$externalCmd($a)" "$matched" result
	    set cmd "| $result"
	    if [catch {set ch [open "$cmd" r]}] {
		return 0
	    }
	    set loadParam(type) channel
	    set loadParam(channel) $ch
	    set loadParam(name) "$fileName"
	    return 1
	}
    }
    return 0
}

# Mosaic File filter procedures

set mosaicTypes {
    mosaicMega
    mosaicGeneric
    mosaicGeneric2
}

proc mosaicGeneric {fileName} {
    if { [regexp {(.*)\_[0-9]+\.fits} $fileName matched root] } {
	return "[glob -nocomplain ${root}_*\.fits ]"
    } else {
	return ""
    }
}

proc mosaicGeneric2 {fileName} {
    if { [regexp {(.*)[0-9]+\.fits} $fileName matched root] } {
	return "[glob -nocomplain ${root}*\.fits ]"
    } else {
	return ""
    }
}

proc mosaicMega { fileName } {
    if { [regexp {(.*)\.IM[0-9]+\.fits} $fileName matched root] } {
	return "[glob -nocomplain ${root}.IM*\.fits]"
    } else {
	return ""
    }
}
