/*
 * Fitsy FITS routines to read and write FITS images.
 */

#include <xos.h>
#include <xfile.h>
#include <fcntl.h>
#include "fitsy.h"

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <sys/ipc.h>
#include <sys/shm.h>

char *strstr();
 
int filesize(int file)
{
  struct stat info;
 
  if (fstat(file, &info) < 0)
    return 0;
  else 
    return info.st_size;
}

int ft_munmap(FITSHead head)
{
  return munmap((caddr_t) head->mmap_here, head->mmap_size);
}

int ft_shmdt(FITSHead head)
{
  return shmdt((caddr_t) head->mmap_here);
}

int ft_simpleimageshm(char* filename, FITSHead* headptr, void** dataptr, 
		      void*** data2dptr, int size, int flags)
{
  char	name[FT_FILENAME];
  char	extn[FT_FILENAME];
  int	indx = 0;

  char	tail[FT_FILENAME];
  int	max = FT_FILENAME;

  int	 shkey;
  char	*data;

  char	*isize;

  ft_parsefilename(filename, name, extn, &indx, tail, max);

  if ( size == 0 && (isize = strstr(tail, "size=")) )
    size = atoi(isize+5);

  /*
   * Attach the memory segment
   */

  shkey = shmget(atoi(name), size, flags);
  data  = (void *) shmat(shkey, 0, 0);

  return ft_indxmemdata(data, size, FT_SHMAT, extn, indx, headptr, dataptr, data2dptr);
} 

int ft_simpleimagemmap(char* filename, FITSHead* headptr, void** dataptr, 
		       void*** data2dptr, int prot, int share)
{
  int file;

  char	name[FT_FILENAME];
  char	extn[FT_FILENAME];
  int	indx = 0;

  char	tail[FT_FILENAME];
  int	max = FT_FILENAME;

  char	*data;
  int	 size;

  ft_parsefilename(filename, name, extn, &indx, tail, max);

  if ((file = open(name, O_RDONLY)) == -1) {
    if (headptr)
      *headptr   = NULL;
    if (data2dptr)
      *data2dptr = NULL;

    return 0;
  }

  /*
   *Map the file.
   */

  size = filesize(file);
  data = (char*)mmap(NULL, size, prot, share, file, 0);
  close(file);

  return ft_indxmemdata(data, size, FT_MMAP, extn, indx, headptr, dataptr, data2dptr);
}

int ft_indxmemdata(char* data, int size, int type, char* extn, int indx, 
		   FITSHead* headptr, void** dataptr, void*** data2dptr)
{
  FITSHead head = NULL;

  char *here = (char *) data;
  void *mapdata = data;
  int mapsize = size;

  /* simple check for fits file */

  if (strncmp(data,"SIMPLE  ",8))
    goto error;

  /* find head and data for specified unit */

  if (*extn) {
    while ((head = ft_headinit(here, size)) && size >= FT_BLOCK) {
      if (!strcmp(extn, ft_name(head)) && (indx == -1 || indx == ft_extn(head)))
	break;
      here += ft_data(head) + ft_databloks(head) * FT_BLOCK;
      size -= ft_data(head) + ft_databloks(head) * FT_BLOCK;
      ft_headfree(head, 0);
    }

    if (!head)
      goto error;
    data = here + ft_data(head);

    /* simple check for valid fits image */

    if (strncmp(head->cards,"XTENSION= 'IMAGE   '",20))
      goto error;
  }
  else if (indx > 0) {
    int	i;
    for (i = 0; i < indx && size >= FT_BLOCK; i++) {
      if (!(head = ft_headinit(here, size)))
	goto error;

      here += ft_data(head) + ft_databloks(head) * FT_BLOCK;
      size -= ft_data(head) + ft_databloks(head) * FT_BLOCK;
      ft_headfree(head, 0);
    }

    if (!(head = ft_headinit(here, size)))
      goto error;
    data = here + ft_data(head);

    /* simple check for valid fits image */

    if (strncmp(head->cards,"XTENSION= 'IMAGE   '",20))
      goto error;
  }
  else {
    if (!(head = ft_headinit(data, size)))
      goto error;
    data += ft_data(head);
  }

  head->mem  = type;
  head->mmap_here = mapdata;
  head->mmap_size = mapsize;

  if (dataptr != NULL)
    *dataptr = data;

  if (data2dptr != NULL) {
    *data2dptr = (void**)ft_make2d((void*)data, Abs(ft_bitpix(head))/8, 
				   0, 0, ft_naxis(head, 1), ft_naxis(head, 2));
  }

  if (headptr != NULL)
    *headptr = head;
  else
    ft_headfree(head, 0);

  return 1;

error:
  if (headptr)
    *headptr = NULL;
  if (data2dptr)
    *data2dptr = NULL;

  if (type == FT_MMAP) 
    munmap(mapdata, mapsize);
  else 
    shmdt(mapdata);

  return 0;
}

int ft_simpleimageread(char* filename, FITSHead* headptr, void** dataptr, 
		       void*** data2dptr, int pixtype)
{
  FITSHead  head = NULL;
  File file;
  int  ret = 1;

  char name[FT_FILENAME];
  char extn[FT_FILENAME];
  int indx = 0;

  char tail[FT_FILENAME];
  int max = FT_FILENAME;

  char initBuf[FT_BLOCK];
  char* buf;
  int size;

  ft_parsefilename(filename, name, extn, &indx, tail, max);

  if (!strcmp(name, "stdin") || !strcmp(name, "STDIN") || !strcmp(name, "-"))
    file = (File)Stdin;
  else if ((file = Open(name, "r")) == NULL)
    goto error;

  /* we need to determin if this is a fits file, so read the first block */

  if (Read(file, initBuf, 1, FT_BLOCK) != FT_BLOCK)
    goto error;

  if (strncmp(initBuf,"SIMPLE  ",8))
    goto error;

  /* ok, now we think we have a fits file, read the rest of the header */

  buf = initBuf;
  size = FT_BLOCK;

  if (*extn) {
    while ((head = ft_headread0(file, buf, size))) {
      if (!strcmp(extn, ft_name(head)) && (indx == -1 || indx == ft_extn(head)))
	break;
      ft_dataskip(file, head, NULL);
      ft_headfree(head, 1);

      /* we want to reassign buf,size to do a full read the next time */
      buf = NULL;
      size = 0;
    }
  }
  else if (indx > 0) {
    int	i;

    for (i = 0; i < indx; i++) {
      head = ft_headread0(file, buf, size);
      if (!head)
	goto error;
      ft_dataskip(file, head, NULL);
      ft_headfree(head, 1);

      /* we want to reassign buf,size to do a full read the next time */
      buf = NULL;
      size = 0;
    }
    head = NULL;
  }

  if (!(ret = ft_imageread(file, &head, dataptr, data2dptr, pixtype, buf, size)))
      goto error;

  if (headptr)
    *headptr = head;
  else
    ft_headfree(head, 1);

  Close(file);

  return 1;

error:
  if (headptr)
    *headptr = NULL;
  if (data2dptr)
    *data2dptr = NULL;

  return 0;
}

int ft_imageread(File file, FITSHead* headptr, void** dataptr, void*** data2dptr,
		 int pixtype, char* buf, int size)
{
  FITSHead fits;
  float *data = NULL;

  if (file == NULL)
    return 0;

  if (headptr == NULL || *headptr == NULL) {
    if ((fits = ft_headread0(file, buf, size)) == NULL)
      return 0;
  } 
  else
    fits = *headptr;

  /* ok, at this point, we can make sure we have a image */

  if (strncmp(fits->cards,"SIMPLE  ",8) && strncmp(fits->cards,"XTENSION= 'IMAGE   '",20))
    return 0;

  if (dataptr)
    data = *dataptr;
  if (data2dptr)
    *data2dptr = NULL;

  if ((data = ft_dataread(file, fits, data, pixtype)) == NULL) {
    if (!headptr)
      ft_headfree(fits, 1);
    return 0;
  }
 
  if (headptr != NULL)
    *headptr = fits;

  if (dataptr != NULL)
    *dataptr = data;

  if (data2dptr != NULL)
    *data2dptr = (void **) ft_make2d((void *) data, Abs(pixtype)/8, 0, 0, 
				     ft_naxis(fits, 1), ft_naxis(fits, 2));

  return 1;
}

int ft_simpleimagewrite(char* filename, FITSHead fits, void* data, int pixtype)
{
  File file;
  FITSHead prim;

  if (fits == NULL)
    return 0;

  file = Open(filename, "wb");

  if (file == NULL) {
    return 0;
  }

  prim = ft_primary(fits);
    
  ft_imagewrite(file, prim, data, pixtype);
  Close(file);

  ft_headfree(prim, 1);
  return 1;
}

int ft_imagewrite(File file, FITSHead fits, void* data, int pixtype)
{
  if (file == NULL)
    return 0;

  ft_syncdata(fits);
  if (ft_headwrite(file, fits) == 0)
    return 0;
  if (ft_datawrite(file, fits, data, pixtype) == 0)
    return 0;

  return 1;
}

