#include <stdlib.h>
#include <string.h>

#include "colorscaletrue24.h"
#include "util.h"

ColorScaleTrueColor24::ColorScaleTrueColor24(int s, Visual* v, int msb)
  : ColorScale(s)
{
  // this makes the assumption of one byte per color.

  colors_ = new unsigned char[s*3];
  int rs = decodeMask((unsigned long)v->red_mask);
  int gs = decodeMask((unsigned long)v->green_mask);
  int bs = decodeMask((unsigned long)v->blue_mask);
      
  // we need to check to byteswap when we have cross platforms

  if ((!msb && lsb()) || (msb && !lsb())) {
    for (int i=0; i<s; i++) {
      unsigned long r = psColors_[i*3+2];
      unsigned long g = psColors_[i*3+1];
      unsigned long b = psColors_[i*3];
      unsigned long a = 0;
      a |= r << rs;
      a |= g << gs;
      a |= b << bs;

      memcpy(colors_+i*3, &a, 3);
    }
  }
  else {
    for (int i=0; i<s; i++) {
      unsigned long r = psColors_[i*3+2];
      unsigned long g = psColors_[i*3+1];
      unsigned long b = psColors_[i*3];
      unsigned long a = 0;
      a |= r << rs;
      a |= g << gs;
      a |= b << bs;

      unsigned char* rr = (unsigned char*)(&a);
      *(colors_+i*3) = *(rr+2);
      *(colors_+i*3+1) = *(rr+1);
      *(colors_+i*3+2) = *(rr);
    }
  }
}

LinearScaleTrueColor24::LinearScaleTrueColor24(int s, 
					       unsigned short* indexCells,
					       unsigned char* colorCells,
					       int count, Visual* v, int msb) 
  : LinearScale(s, indexCells, colorCells, count), 
    ColorScaleTrueColor24(s, v, msb), 
    ColorScale(s) {}

LogScaleTrueColor24::LogScaleTrueColor24(int s,
					 unsigned short* indexCells,
					 unsigned char* colorCells,
					 int count, Visual* v, int msb)
  : LogScale(s, indexCells, colorCells, count), 
    ColorScaleTrueColor24(s, v, msb), 
    ColorScale(s) {}

SquaredScaleTrueColor24::SquaredScaleTrueColor24(int s,
						 unsigned short* indexCells,
						 unsigned char* colorCells,
						 int count, Visual* v, int msb)
  : SquaredScale(s, indexCells, colorCells, count), 
    ColorScaleTrueColor24(s, v, msb), 
    ColorScale(s) {}

SqrtScaleTrueColor24::SqrtScaleTrueColor24(int s,
					   unsigned short* indexCells,
					   unsigned char* colorCells,
					   int count, Visual* v, int msb)
  : SqrtScale(s, indexCells, colorCells, count), 
    ColorScaleTrueColor24(s, v, msb), 
    ColorScale(s) {}

IISScaleTrueColor24::IISScaleTrueColor24(unsigned short* indexCells,
					 unsigned char* colorCells,
					 int count, Visual* v, int msb) 
  : IISScale(indexCells, colorCells, count),
    ColorScaleTrueColor24(IISSIZE, v, msb), 
    ColorScale(IISSIZE) {}
